/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * @file include/util/semaphore.h
 **/

#include "config.h"
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>

#ifndef SEMAPHORES_H
#define SEMAPHORES_H

/* do not turn this on unless you know what you
   are doing, you'll get a ton of output... */
#define DEBUG_LOCKING 0

typedef struct Mutex {
  pthread_mutex_t pthreadMutex;
} Mutex;

typedef struct Semaphore {
  int v;
  Mutex mutex;
  pthread_cond_t cond;
} Semaphore;

/* ************** deprecated operations ************* */

Semaphore * new_semaphore (int value);
void free_semaphore(Semaphore * s);
int semaphore_down (Semaphore * s);
int semaphore_up (Semaphore * s);

/* ************** Semaphore operations ************* */

#define SEMAPHORE_NEW(value) semaphore_new_(value, __FILE__, __LINE__)
#define SEMAPHORE_FREE(s) semaphore_free_(s, __FILE__, __LINE__)
#define SEMAPHORE_DOWN(s) semaphore_down_(s, __FILE__, __LINE__)
#define SEMAPHORE_UP(s) semaphore_up_(s, __FILE__, __LINE__)

/* ************** MUTEX OPERATIONS ***************** */
#if DEBUG_LOCKING
#define MUTEX_CREATE(a) do { \
  fprintf(stderr, \
          "Creating mutex %x at line %d in file %s\n", \
          (int) a, __LINE__, __FILE__); \
  create_mutex_(a); \
}\
while(0)
#define MUTEX_CREATE_RECURSIVE(a) do { \
  fprintf(stderr, \
          "Creating recursive mutex %x at line %d in file %s\n", \
          (int) a, __LINE__, __FILE__); \
  create_recursive_mutex_(a); \
}\
while(0)
#define MUTEX_DESTROY(a) do { \
  fprintf(stderr, \
          "Destroying mutex %x at line %d in file %s\n", \
          (int) a, __LINE__, __FILE__); \
  destroy_mutex_(a); \
}\
while(0)
#define MUTEX_LOCK(a) do { \
  fprintf(stderr, \
          "Aquireing lock %x at line %d in file %s\n", \
          (int)a, __LINE__, __FILE__); \
  mutex_lock_(a); \
}\
while (0)
#define MUTEX_UNLOCK(a) do { \
  fprintf(stderr, \
         "Releasing lock %x at line %d in file %s\n", \
	(int)a,__LINE__, __FILE__); \
  mutex_unlock_(a); \
}\
while (0)
#else
#define MUTEX_LOCK(a) mutex_lock_(a)
#define MUTEX_UNLOCK(a) mutex_unlock_(a)
#define MUTEX_CREATE(a) create_mutex_(a)
#define MUTEX_CREATE_RECURSIVE(a) create_recursive_mutex_(a)
#define MUTEX_DESTROY(a) destroy_mutex_(a)
#endif


/* *************** thread handling ************* */

int PTHREAD_CREATE(pthread_t * handle,
		   void * (*main)(void *),
		   void * arg,
		   size_t stackSize);



/* ***************** internal methods ************* */

/**
 * While we must define these globally to make the
 * compiler happy, always use the macros in the sources
 * instead! 
 **/
void create_mutex_(Mutex * mutex);
void create_recursive_mutex_(Mutex * mutex);
void create_fast_mutex_(Mutex * mutex);
void destroy_mutex_(Mutex * mutex);
void mutex_lock_(Mutex * mutex);
void mutex_unlock_(Mutex * mutex);
Semaphore * semaphore_new_(int value, 
			   const char * filename,
			   const int linenumber);
void semaphore_free_(Semaphore * s,
		     const char * filename,
		     const int linenumber);
int semaphore_down_(Semaphore * s,
		    const char * filename,
		    const int linenumber);
int semaphore_up_(Semaphore * s,
		  const char * filename,
		  const int linenumber);		    

#endif
